import inspect
import logging

from scrapy.spiders import Spider
from scrapy.utils.defer import deferred_from_coro
from scrapy.utils.misc import arg_to_iter
from scrapy.settings import register_settings
try:
    from scrapy.utils.py36 import collect_asyncgen
except SyntaxError:
    collect_asyncgen = None


logger = logging.getLogger(__name__)


def iterate_spider_output(result):
    if collect_asyncgen and hasattr(inspect, 'isasyncgen') and inspect.isasyncgen(result):
        d = deferred_from_coro(collect_asyncgen(result))
        d.addCallback(iterate_spider_output)
        return d
    elif inspect.iscoroutine(result):
        d = deferred_from_coro(result)
        d.addCallback(iterate_spider_output)
        return d
    return arg_to_iter(result)


def iter_spider_classes(module):
    """
    by zhizhong
    Return an iterator over all spider classes defined in the given module
    that can be instantiated (i.e. which have name)
    """
    # this needs to be imported here until get rid of the spider manager
    # singleton in scrapy.spider.spiders
    from scrapy.spiders import Spider
    try:
        import spider_register
    except:
        logging.warning("注意检查!! 项目中是否存在 spider_register 注册文件")
    for obj in vars(module).values():
        if inspect.isclass(obj) and len(obj.__module__.split("."))==5:
            project_dir, spider_path, user_name, register_key = obj.__module__.split(".",3)
            if (
                issubclass(obj, Spider)
                and obj.__module__ == module.__name__
                ):
                try:
                    if getattr(obj, 'name') is None or isinstance(obj.name,list):
                        obj.name = eval(f"spider_register.{user_name}.get('{register_key}')[0]") # to_fixed
                        obj = filter_settings_by_register(obj,eval(f"spider_register.{user_name}.get('{register_key}')"))
                except Exception as e: 
                    print(f"经过检查,{user_name}同学的爬虫模块{register_key}名为{obj.name}的配置存在匹配异常：{e}")
                yield obj

def filter_settings_by_register(obj,param):
    """
    by zhizhong
    根据spider_register 文件中管道中间件配置情况过滤爬虫设置
    """
    relate_dic = {
        "stickyitem":"Scrapy_Prj.middlewares.spidermiddlewares.itemheaders.ItemHeadersMiddleware",
        "ua":"Scrapy_Prj.middlewares.downloadermiddlewares.randuamiddleware.RandomUserAgent",
        "proxy":"Scrapy_Prj.middlewares.downloadermiddlewares.proxymiddleware.ProxyMiddleware",
        "file":"Scrapy_Prj.pipelines.ossfilepipelines.OssFilesPipeline",
        "file_bak":"Scrapy_Prj.pipelines.ossfilepipelines.OssFilesPipelineBak",
        "img":"Scrapy_Prj.pipelines.ossimgpipelines.OssImagesPipeline",
        "img_bak":"Scrapy_Prj.pipelines.ossimgpipelines.OssImagesPipelineBak",
        "pyppeteer":"gerapy_pyppeteer.downloadermiddlewares.PyppeteerMiddleware",
        "bloom":"Scrapy_Prj.pipelines.bloomerfiler.BloomFilterPipeline",
        "mysql":"Scrapy_Prj.pipelines.mysqlpipelines.MySQLPipeline",
        "mongo":"Scrapy_Prj.pipelines.mongodbPipelines.MongodbPipeline",
        "rabbit":"Scrapy_Prj.pipelines.rabbitmqpipelines.RabbitMQPipeline",
        "kafka":"Scrapy_Prj.pipelines.kafkaPipelines.KafkaPipeline",
        "influxdb":"Scrapy_Prj.pipelines.influxdbpipelines.SpiderStatLogging",
        "closespider":"scrapy.extensions.closespider.CloseSpider",
        "fctexcel":"Scrapy_Prj.pipelines.excelpipelines.ExcelFactoryPipeline",
        "dpfexcel":"Scrapy_Prj.pipelines.excelpipelines.ExcelDataplatformPipeline",
        "mailer":"scrapy.extensions.extensions.statsmailer.StatsMailer",
    }
    obj.custom_settings =obj.custom_settings if obj.custom_settings else {}
    # 为防止重复引用,对custom_settings引用源码中scrapy.xcc_管道或中间件置空
    if set(i for i in obj.custom_settings.get("SPIDER_MIDDLEWARES",{}).keys()if i.startswith("scrapy.xcc_")):
        obj.custom_settings["SPIDER_MIDDLEWARES"] = {}
    if set(i for i in obj.custom_settings.get("DOWNLOADER_MIDDLEWARES",{}).keys()if i.startswith("scrapy.xcc_")):
        obj.custom_settings["DOWNLOADER_MIDDLEWARES"] = {}
    if set(i for i in obj.custom_settings.get("ITEM_PIPELINES",{}).keys()if i.startswith("scrapy.xcc_")):
        obj.custom_settings["ITEM_PIPELINES"] = {}
    if set(i for i in obj.custom_settings.get("EXTENSIONS",{}).keys()if i.startswith("scrapy.xcc_")):
        obj.custom_settings["EXTENSIONS"] = {}
    # 取spider_register中开启管道
    obj.custom_settings = dict() if obj.custom_settings is None else obj.custom_settings
    settings_temp = register_settings
    spider_middlewares = {relate_dic[i]  for i in param[1] if isinstance(param[1],tuple)} or {relate_dic[param[1]] for i in param[1] if param[1]} 
    downloader_middlewares = {relate_dic[i]  for i in param[2] if isinstance(param[2],tuple)} or {relate_dic[param[2]] for i in param[2] if param[2]} 
    pipelines = {relate_dic[i]  for i in param[3] if isinstance(param[3],tuple)} or {relate_dic[param[3]] for i in param[3] if param[3]} 
    extensions ={relate_dic[i]  for i in param[4] if isinstance(param[4],tuple)} or {relate_dic[param[4]] for i in param[4] if param[4]} 

    temp_dict = dict()
    temp_dict["SPIDER_MIDDLEWARES"] = obj.custom_settings.get("SPIDER_MIDDLEWARES",{})
    temp_dict["SPIDER_MIDDLEWARES"].update({k:v for k,v in settings_temp.SPIDER_MIDDLEWARES.items() if k in spider_middlewares})
    temp_dict["DOWNLOADER_MIDDLEWARES"] = obj.custom_settings.get("DOWNLOADER_MIDDLEWARES",{})
    temp_dict["DOWNLOADER_MIDDLEWARES"].update({k:v for k,v in settings_temp.DOWNLOADER_MIDDLEWARES.items() if k in downloader_middlewares})
    temp_dict["ITEM_PIPELINES"] = obj.custom_settings.get("ITEM_PIPELINES",{})
    temp_dict["ITEM_PIPELINES"].update({k:v for k,v in settings_temp.ITEM_PIPELINES.items() if k in pipelines})
    temp_dict["EXTENSIONS"] = obj.custom_settings.get("EXTENSIONS",{})
    temp_dict["EXTENSIONS"].update({k:v for k,v in settings_temp.EXTENSIONS.items() if k in extensions})
    obj.custom_settings.update(temp_dict)
    return obj


def spidercls_for_request(spider_loader, request, default_spidercls=None,
                          log_none=False, log_multiple=False):
    """Return a spider class that handles the given Request.

    This will look for the spiders that can handle the given request (using
    the spider loader) and return a Spider class if (and only if) there is
    only one Spider able to handle the Request.

    If multiple spiders (or no spider) are found, it will return the
    default_spidercls passed. It can optionally log if multiple or no spiders
    are found.
    """
    snames = spider_loader.find_by_request(request)
    if len(snames) == 1:
        return spider_loader.load(snames[0])

    if len(snames) > 1 and log_multiple:
        logger.error('More than one spider can handle: %(request)s - %(snames)s',
                     {'request': request, 'snames': ', '.join(snames)})

    if len(snames) == 0 and log_none:
        logger.error('Unable to find spider that handles: %(request)s',
                     {'request': request})

    return default_spidercls


class DefaultSpider(Spider):
    name = 'default'



# -*- coding: utf-8 -*-
from scrapy.utils.conf import get_config
import pika
import logging

     
from twisted.internet.threads import deferToThread
from scrapy.utils.serialize import ScrapyJSONEncoder
import time,os

EXCHANGE_NAME = ''

import logging
logger = logging.getLogger('pika').setLevel(logging.WARNING)
class RabbitMQPipeline(object):
    """Pushes serialized item into a RabbitMQ list/queue"""

    def __init__(self, item_key, connection_url):
        self.item_key = item_key
        self.exchange_name = EXCHANGE_NAME
        self.encoder = ScrapyJSONEncoder()
        self.connection_url= connection_url

    @classmethod
    def from_crawler(cls, crawler):
        section_select ="mq_cfg_prod" if os.environ.get('IF_PROD') == "True" or crawler.settings.get('IF_PROD')\
             == True or get_config().get("settings","IF_PROD",fallback="False")=="True" else "mq_cfg_dev"
        if hasattr(crawler.spider, 'items_key'):
            item_key = crawler.spider.items_key
        elif get_config().get(section=section_select,option="RABBITMQ_ITEMS_KEY",fallback=''):
            item_key = get_config().get(section=section_select,option="RABBITMQ_ITEMS_KEY",fallback='')
        elif hasattr(crawler.settings,"RABBITMQ_ITEMS_KEY"):
            item_key = crawler.settings.get("RABBITMQ_ITEMS_KEY")
        else:
            item_key = 'items_{spider_name}'.format(
                spider_name=crawler.spider.name)
        connection_url =  get_config().get(section=section_select,option="RABBITMQ_ITEMS_KEY",fallback='') or \
            crawler.settings.get('RABBITMQ_ITEMS_KEY')
        return cls(item_key,connection_url)

    def process_item(self, item, spider):
        return deferToThread(self._process_item, item, spider)

    def _process_item(self, item, spider):
        key = self.item_key
        data = self.encoder.encode(item)
        try_time = 1
        while try_time<10:
            try:
                self.server = self.connect_mq(self.connection_url)
                self.channel = self.get_channel(self.server, self.item_key)
                self.channel.basic_publish(exchange=self.exchange_name,
                                        routing_key=key,
                                        body=data)
                return item
            except Exception as e:
                logger.exception(e)
                logger.error('process item failed! try_time:{}'.format(try_time))
                try_time += 1
                time.sleep(1)
                self.channel = self.get_channel(self.server, self.item_key)
        return item

    def get_channel(self,connection, queue_name, durable=True, confirm_delivery=True, is_delay=False):
        """ Init method to return a prepared channel for consuming
        """
        channel = connection.channel()
        channel.queue_declare(queue=queue_name,durable=durable)#,arguments={'x-max-priority': 255,'vhost':'/'})
        if confirm_delivery:
            channel.confirm_delivery()

        if is_delay is True:
            exchange_name = "{}-delay".format(queue_name)
            channel.exchange_declare(exchange_name,
                                    exchange_type="x-delayed-message",
                                    arguments={"x-delayed-type": "direct"})
            channel.queue_bind(
                queue=queue_name, exchange=exchange_name, routing_key=queue_name)
        return channel
    
    def connect_mq(self,connection_url):
        """ Create and return a fresh connection
        """
        return pika.BlockingConnection(pika.URLParameters(connection_url))



import hashlib
import tempfile
import unittest
import shutil
import os
import sys
from testfixtures import LogCapture

from scrapy.dupefilters import RFPDupeFilter
from scrapy.http import Request
from scrapy.core.scheduler import Scheduler
from scrapy.utils.python import to_bytes
from scrapy.utils.job import job_dir
from scrapy.utils.test import get_crawler
from tests.spiders import SimpleSpider


class FromCrawlerRFPDupeFilter(RFPDupeFilter):

    @classmethod
    def from_crawler(cls, crawler):
        debug = crawler.settings.getbool('DUPEFILTER_DEBUG')
        df = cls(job_dir(crawler.settings), debug)
        df.method = 'from_crawler'
        return df


class FromSettingsRFPDupeFilter(RFPDupeFilter):

    @classmethod
    def from_settings(cls, settings):
        debug = settings.getbool('DUPEFILTER_DEBUG')
        df = cls(job_dir(settings), debug)
        df.method = 'from_settings'
        return df


class DirectDupeFilter:
    method = 'n/a'


class RFPDupeFilterTest(unittest.TestCase):

    def test_df_from_crawler_scheduler(self):
        settings = {'DUPEFILTER_DEBUG': True,
                    'DUPEFILTER_CLASS': FromCrawlerRFPDupeFilter}
        crawler = get_crawler(settings_dict=settings)
        scheduler = Scheduler.from_crawler(crawler)
        self.assertTrue(scheduler.df.debug)
        self.assertEqual(scheduler.df.method, 'from_crawler')

    def test_df_from_settings_scheduler(self):
        settings = {'DUPEFILTER_DEBUG': True,
                    'DUPEFILTER_CLASS': FromSettingsRFPDupeFilter}
        crawler = get_crawler(settings_dict=settings)
        scheduler = Scheduler.from_crawler(crawler)
        self.assertTrue(scheduler.df.debug)
        self.assertEqual(scheduler.df.method, 'from_settings')

    def test_df_direct_scheduler(self):
        settings = {'DUPEFILTER_CLASS': DirectDupeFilter}
        crawler = get_crawler(settings_dict=settings)
        scheduler = Scheduler.from_crawler(crawler)
        self.assertEqual(scheduler.df.method, 'n/a')

    def test_filter(self):
        dupefilter = RFPDupeFilter()
        dupefilter.open()

        r1 = Request('http://scrapytest.org/1')
        r2 = Request('http://scrapytest.org/2')
        r3 = Request('http://scrapytest.org/2')

        assert not dupefilter.request_seen(r1)
        assert dupefilter.request_seen(r1)

        assert not dupefilter.request_seen(r2)
        assert dupefilter.request_seen(r3)

        dupefilter.close('finished')

    def test_dupefilter_path(self):
        r1 = Request('http://scrapytest.org/1')
        r2 = Request('http://scrapytest.org/2')

        path = tempfile.mkdtemp()
        try:
            df = RFPDupeFilter(path)
            try:
                df.open()
                assert not df.request_seen(r1)
                assert df.request_seen(r1)
            finally:
                df.close('finished')

            df2 = RFPDupeFilter(path)
            try:
                df2.open()
                assert df2.request_seen(r1)
                assert not df2.request_seen(r2)
                assert df2.request_seen(r2)
            finally:
                df2.close('finished')
        finally:
            shutil.rmtree(path)

    def test_request_fingerprint(self):
        """Test if customization of request_fingerprint method will change
        output of request_seen.

        """
        r1 = Request('http://scrapytest.org/index.html')
        r2 = Request('http://scrapytest.org/INDEX.html')

        dupefilter = RFPDupeFilter()
        dupefilter.open()

        assert not dupefilter.request_seen(r1)
        assert not dupefilter.request_seen(r2)

        dupefilter.close('finished')

        class CaseInsensitiveRFPDupeFilter(RFPDupeFilter):

            def request_fingerprint(self, request):
                fp = hashlib.sha1()
                fp.update(to_bytes(request.url.lower()))
                return fp.hexdigest()

        case_insensitive_dupefilter = CaseInsensitiveRFPDupeFilter()
        case_insensitive_dupefilter.open()

        assert not case_insensitive_dupefilter.request_seen(r1)
        assert case_insensitive_dupefilter.request_seen(r2)

        case_insensitive_dupefilter.close('finished')

    def test_seenreq_newlines(self):
        """ Checks against adding duplicate \r to
        line endings on Windows platforms. """

        r1 = Request('http://scrapytest.org/1')

        path = tempfile.mkdtemp()
        try:
            df = RFPDupeFilter(path)
            df.open()
            df.request_seen(r1)
            df.close('finished')

            with open(os.path.join(path, 'requests.seen'), 'rb') as seen_file:
                line = next(seen_file).decode()
                assert not line.endswith('\r\r\n')
                if sys.platform == 'win32':
                    assert line.endswith('\r\n')
                else:
                    assert line.endswith('\n')

        finally:
            shutil.rmtree(path)

    def test_log(self):
        with LogCapture() as log:
            settings = {'DUPEFILTER_DEBUG': False,
                        'DUPEFILTER_CLASS': FromCrawlerRFPDupeFilter}
            crawler = get_crawler(SimpleSpider, settings_dict=settings)
            scheduler = Scheduler.from_crawler(crawler)
            spider = SimpleSpider.from_crawler(crawler)

            dupefilter = scheduler.df
            dupefilter.open()

            r1 = Request('http://scrapytest.org/index.html')
            r2 = Request('http://scrapytest.org/index.html')

            dupefilter.log(r1, spider)
            dupefilter.log(r2, spider)

            assert crawler.stats.get_value('dupefilter/filtered') == 2
            log.check_present(
                (
                    'scrapy.dupefilters',
                    'DEBUG',
                    'Filtered duplicate request: <GET http://scrapytest.org/index.html> - no more'
                    ' duplicates will be shown (see DUPEFILTER_DEBUG to show all duplicates)'
                )
            )

            dupefilter.close('finished')

    def test_log_debug(self):
        with LogCapture() as log:
            settings = {'DUPEFILTER_DEBUG': True,
                        'DUPEFILTER_CLASS': FromCrawlerRFPDupeFilter}
            crawler = get_crawler(SimpleSpider, settings_dict=settings)
            scheduler = Scheduler.from_crawler(crawler)
            spider = SimpleSpider.from_crawler(crawler)

            dupefilter = scheduler.df
            dupefilter.open()

            r1 = Request('http://scrapytest.org/index.html')
            r2 = Request('http://scrapytest.org/index.html',
                         headers={'Referer': 'http://scrapytest.org/INDEX.html'})

            dupefilter.log(r1, spider)
            dupefilter.log(r2, spider)

            assert crawler.stats.get_value('dupefilter/filtered') == 2
            log.check_present(
                (
                    'scrapy.dupefilters',
                    'DEBUG',
                    'Filtered duplicate request: <GET http://scrapytest.org/index.html> (referer: None)'
                )
            )
            log.check_present(
                (
                    'scrapy.dupefilters',
                    'DEBUG',
                    'Filtered duplicate request: <GET http://scrapytest.org/index.html>'
                    ' (referer: http://scrapytest.org/INDEX.html)'
                )
            )

            dupefilter.close('finished')
